/******************************************************************************
*
*       Author:  K. Chaudhary / S. Nag / H. Varma
*      Company:  Xilinx
*
*   Disclaimer:  THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY 
*                WHATSOEVER AND XILINX SPECIFICALLY DISCLAIMS ANY 
*                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
*                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
*
*                Copyright (c) 1999 Xilinx, Inc.
*                All rights reserved
*
*
******************************************************************************/

`timescale 1ns / 1ps



/*****************************************************************************
* ByteMultOdd implements the following
*    dout0 = (`C1*din0+`C3*din1+`C5*din2+`C7*din3);
*    dout1 = (`C3*din0-`C7*din1-`C1*din2-`C5*din3);
*    dout2 = (`C5*din0-`C1*din1+`C7*din2+`C3*din3);
*    dout3 = (`C7*din0-`C5*din1+`C3*din2-`C1*din3);
*
*    Inputs are 12 bits long and the outputs are 15 bits long to increase the 
*    dynamic range of the numbers.
*
*    This computation is being done using Distributed Arithemetic (DALUTs) in
*    bit serial mode. See App Note reference for more details.
*
******************************************************************************/

module ByteMultOdd(dout0, dout1, dout2, dout3, din0, din1, din2, din3, clock, go , rst, done);

`define width 14

// shift the input din by 5 bits and sign extend the number by two bits
// to produce 15 bit number.
function [14:0] sgnExtend;
	input [7:0] din;
	sgnExtend = {din[7],din[7], din, 5'b0};	
endfunction

input [11:0] din0;
input [11:0] din1;
input [11:0] din2;
input [11:0] din3;
input clock;
input go;
input rst;
output [`width:0] dout0;
wire [`width:0] dout0;
output [`width:0] dout1;
wire [`width:0] dout1;
output [`width:0] dout2;
wire [`width:0] dout2;
output [`width:0] dout3;
wire [`width:0] dout3;
output done;
reg done;
wire [7:0] d0;
wire [7:0] d1;
wire [7:0] d2;
wire [7:0] d3;

reg [3:0] count;
reg compute;
reg [11:0] in0, in1, in2, in3;

// internal registers
reg [`width:0] int_dout0;
reg [`width:0] int_dout1;
reg [`width:0] int_dout2;
reg [`width:0] int_dout3;

assign dout0 = int_dout0;
assign dout1 = int_dout1;
assign dout2 = int_dout2;
assign dout3 = int_dout3;

biglut1 struct0 (d0, in0[0], in1[0], in2[0], in3[0]);
biglut3 struct1 (d1, in0[0], in1[0], in2[0], in3[0]);
biglut5 struct2 (d2, in0[0], in1[0], in2[0], in3[0]);
biglut7 struct3 (d3, in0[0], in1[0], in2[0], in3[0]);

always @(posedge clock or posedge rst)
begin
  if (rst == 1'b1) begin
    	count = 4'b0;
    	int_dout0 = `width+1'b0;
    	int_dout1 = `width+1'b0;
	int_dout2 = `width+1'b0;
	int_dout3 = `width+1'b0;
    	done = 1'b0;
	compute = 1'b0;
    	in0 = 12'b0;
    	in1 = 12'b0;
    	in2 = 12'b0;
    	in3 = 12'b0;
  end
  else begin

     if (done == 1'b1)
	done = 1'b0;

    // Accumulate Block
	if (compute == 1'b1) begin
	  if (count == 4'd11) begin
        	int_dout0 = {int_dout0[`width],int_dout0[`width:1]} - sgnExtend(d0);
		int_dout1 = {int_dout1[`width],int_dout1[`width:1]} - sgnExtend(d1);
		int_dout2 = {int_dout2[`width],int_dout2[`width:1]} - sgnExtend(d2);
		int_dout3 = {int_dout3[`width],int_dout3[`width:1]} - sgnExtend(d3);
        	done = 1'b1;
        	compute = 1'b0;
	  end
      else begin
        	int_dout0 = {int_dout0[`width],int_dout0[`width:1]} + sgnExtend(d0);
		int_dout1 = {int_dout1[`width],int_dout1[`width:1]} + sgnExtend(d1);
		int_dout2 = {int_dout2[`width],int_dout2[`width:1]} + sgnExtend(d2);
		int_dout3 = {int_dout3[`width],int_dout3[`width:1]} + sgnExtend(d3);
      end
      count = count+1;
    end

    // compute block
    if ((go == 1'b1) && (done == 1'b0))
    	  compute = 1'b1;

    // shift inputs
    if (go == 1'b1) begin
      in0 = din0;
      in1 = din1;
      in2 = din2;
      in3 = din3;
      count = 4'b0;
      int_dout0 = `width+1'b0;
      int_dout1 = `width+1'b0;
      int_dout2 = `width+1'b0;
      int_dout3 = `width+1'b0;      	
      end
    else begin
      in0[10:0] = in0[11:1];
      in1[10:0] = in1[11:1];
      in2[10:0] = in2[11:1];
      in3[10:0] = in3[11:1];
    end

  end // else
end

endmodule


/*****************************************************************************
* ByteMultEven implements the following
*    dout0 = (`C0*din0+`C2*din1+`C4*din2+`C6*din3);
*    dout1 = (`C0*din0+`C6*din1-`C4*din2-`C2*din3);
*    dout2 = (`C0*din0-`C6*din1-`C4*din2+`C2*din3);
*    dout3 = (`C0*din0-`C2*din1+`C4*din2-`C6*din3);
*
*    Inputs are 12 bits long and the outputs are 15 bits long to increase the 
*    dynamic range of the numbers.
*
*    This computation is being done using Distributed Arithemetic (DALUTs) in 
*    bit serial mode. See App Note reference for more details.
*
******************************************************************************/

module ByteMultEven(dout0, dout1, dout2, dout3, din0, din1, din2, din3, clock, go , rst, done);

`define width 14

// shift the input din by 5 bits and sign extend the number by two bits
// to produce 15 bit number.
function [14:0] sgnExtend;
	input [7:0] din;
	sgnExtend = {din[7],din[7],din, 5'b0};	
endfunction

input [11:0] din0;
input [11:0] din1;
input [11:0] din2;
input [11:0] din3;
input clock;
input go;
input rst;
output [`width:0] dout0;
wire [`width:0] dout0;
output [`width:0] dout1;
wire [`width:0] dout1;
output [`width:0] dout2;
wire [`width:0] dout2;
output [`width:0] dout3;
wire [`width:0] dout3;
output done;
reg done;
wire [7:0] d0;
wire [7:0] d1;
wire [7:0] d2;
wire [7:0] d3;

reg [3:0] count;
reg compute;
reg [11:0] in0, in1, in2, in3;

// internal registers
reg [`width:0] int_dout0;
reg [`width:0] int_dout1;
reg [`width:0] int_dout2;
reg [`width:0] int_dout3;

assign dout0 = int_dout0;
assign dout1 = int_dout1;
assign dout2 = int_dout2;
assign dout3 = int_dout3;

biglut0 struct0 (d0, in0[0], in1[0], in2[0], in3[0]);
biglut2 struct1 (d1, in0[0], in1[0], in2[0], in3[0]);
biglut4 struct2 (d2, in0[0], in1[0], in2[0], in3[0]);
biglut6 struct3 (d3, in0[0], in1[0], in2[0], in3[0]);

always @(posedge clock or posedge rst)
begin
  if (rst == 1'b1) begin
    	count = 4'b0;
    	int_dout0 = `width+1'b0;
    	int_dout1 = `width+1'b0;
	int_dout2 = `width+1'b0;
	int_dout3 = `width+1'b0;
    	done = 1'b0;
	compute = 1'b0;
    	in0 = 12'b0;
    	in1 = 12'b0;
    	in2 = 12'b0;
    	in3 = 12'b0;
  end
  else begin

     if (done == 1'b1)
	done = 1'b0;

    // Accumulate Block
	if (compute == 1'b1) begin
	  if (count == 4'd11) begin
        	int_dout0 = {int_dout0[`width],int_dout0[`width:1]} - sgnExtend(d0);
		int_dout1 = {int_dout1[`width],int_dout1[`width:1]} - sgnExtend(d1);
		int_dout2 = {int_dout2[`width],int_dout2[`width:1]} - sgnExtend(d2);
		int_dout3 = {int_dout3[`width],int_dout3[`width:1]} - sgnExtend(d3);
        	done = 1'b1;
        	compute = 1'b0;
	  end
      else begin
        	int_dout0 = {int_dout0[`width],int_dout0[`width:1]} + sgnExtend(d0);
		int_dout1 = {int_dout1[`width],int_dout1[`width:1]} + sgnExtend(d1);
		int_dout2 = {int_dout2[`width],int_dout2[`width:1]} + sgnExtend(d2);
		int_dout3 = {int_dout3[`width],int_dout3[`width:1]} + sgnExtend(d3);
      end
      count = count+1;
    end

    // compute block
    if ((go == 1'b1) && (done == 1'b0))
    	  compute = 1'b1;

    // shift inputs
    if (go == 1'b1) begin
      in0 = din0;
      in1 = din1;
      in2 = din2;
      in3 = din3;
      count = 4'b0;
      int_dout0 = `width+1'b0;
      int_dout1 = `width+1'b0;
      int_dout2 = `width+1'b0;
      int_dout3 = `width+1'b0;      	
      end
    else begin
      in0[10:0] = in0[11:1];
      in1[10:0] = in1[11:1];
      in2[10:0] = in2[11:1];
      in3[10:0] = in3[11:1];
    end

  end // else
end

endmodule

