/******************************************************************************
*
*       Author:  K. Chaudhary / S. Nag / H. Varma
*      Company:  Xilinx
*
*   Disclaimer:  THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY 
*                WHATSOEVER AND XILINX SPECIFICALLY DISCLAIMS ANY 
*                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
*                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
*
*                Copyright (c) 1999 Xilinx, Inc.
*                All rights reserved
*
*
******************************************************************************/

`timescale 1ns / 1ps

/******************************************************************************
*
* implements 2-D IDCT 
* This module takes in a matrix of input frequency, then multilies it by
* the matrix of IDCT coefficients and returns a row of output discrete time
* values.
*         I0, I1, I2 ... I7 are input frequencies.
*         O0, O1, O2 ... O7 are output discrete time values.
* 
* input is stored as
*    din[0:95]    = D(0,0)[0] .. D(0,0)[11] D(0,1)[0]..D(0,1)[11] ... D(0,7)[0]..D(0,7)[11]
*    din[96:191]  = D(1,0)[0] .. D(1,0)[11] D(1,1)[0]..D(1,1)[11] ... D(1,7)[0]..D(1,7)[11]
*    ....
*    din[672:767] = D(7,0)[0] .. D(7,0)[11] D(7,1)[0]..D(7,1)[11] ... D(7,7)[0]..D(7,7)[11]
*
******************************************************************************/
module idct(din, dout, clk, go, rst, done);

`define bussize 767

input [0:`bussize] din;
input clk;
input go;
input rst;
output done;
reg done;
output [0:`bussize] dout;
reg  [0:`bussize] dout;
wire [0:`bussize] dout1D;
wire [0:`bussize] dout2D;
wire  donerow, donecol;
wire doneC1, doneC2, doneC3, doneC4, doneC5, doneC6, doneC7;
wire doneR1, doneR2, doneR3, doneR4, doneR5, doneR6, doneR7;


// instantiate 8 one_d_idct for row computations
// 1D DCT row-in (din) -> 1D-DCT row out (dout1D)

    one_d_idct UR0 (din[0:11], din[12:23], din[24:35], din[36:47], din[48:59], din[60:71], din[72:83], din[84:95], dout1D[0:11], dout1D[12:23], dout1D[24:35], dout1D[36:47], dout1D[48:59], dout1D[60:71], dout1D[72:83], dout1D[84:95], clk, go, rst, donerow);
    one_d_idct UR1 (din[96:107], din[108:119], din[120:131], din[132:143], din[144:155], din[156:167], din[168:179], din[180:191], dout1D[96:107], dout1D[108:119], dout1D[120:131], dout1D[132:143], dout1D[144:155], dout1D[156:167], dout1D[168:179], dout1D[180:191], clk, go, rst, doneR1);
    one_d_idct UR2 (din[192:203], din[204:215], din[216:227], din[228:239], din[240:251], din[252:263], din[264:275], din[276:287], dout1D[192:203], dout1D[204:215], dout1D[216:227], dout1D[228:239], dout1D[240:251], dout1D[252:263], dout1D[264:275], dout1D[276:287], clk, go, rst, doneR2);
    one_d_idct UR3 (din[288:299], din[300:311], din[312:323], din[324:335], din[336:347], din[348:359], din[360:371], din[372:383], dout1D[288:299], dout1D[300:311], dout1D[312:323], dout1D[324:335], dout1D[336:347], dout1D[348:359], dout1D[360:371], dout1D[372:383], clk, go, rst, doneR3);
    one_d_idct UR4 (din[384:395], din[396:407], din[408:419], din[420:431], din[432:443], din[444:455], din[456:467], din[468:479], dout1D[384:395], dout1D[396:407], dout1D[408:419], dout1D[420:431], dout1D[432:443], dout1D[444:455], dout1D[456:467], dout1D[468:479], clk, go, rst, doneR4);
    one_d_idct UR5 (din[480:491], din[492:503], din[504:515], din[516:527], din[528:539], din[540:551], din[552:563], din[564:575], dout1D[480:491], dout1D[492:503], dout1D[504:515], dout1D[516:527], dout1D[528:539], dout1D[540:551], dout1D[552:563], dout1D[564:575], clk, go, rst, doneR5);
    one_d_idct UR6 (din[576:587], din[588:599], din[600:611], din[612:623], din[624:635], din[636:647], din[648:659], din[660:671], dout1D[576:587], dout1D[588:599], dout1D[600:611], dout1D[612:623], dout1D[624:635], dout1D[636:647], dout1D[648:659], dout1D[660:671], clk, go, rst, doneR6);
    one_d_idct UR7 (din[672:683], din[684:695], din[696:707], din[708:719], din[720:731], din[732:743], din[744:755], din[756:767], dout1D[672:683], dout1D[684:695], dout1D[696:707], dout1D[708:719], dout1D[720:731], dout1D[732:743], dout1D[744:755], dout1D[756:767], clk, go, rst, doneR7);


// instantiate 8 one_d_idct for column computations
// 1D DCT col-in (dout1D) -> 1D-DCT col out (dout)

    one_d_idct UC0 (dout1D[0:11], dout1D[96:107], dout1D[192:203], dout1D[288:299], dout1D[384:395], dout1D[480:491], dout1D[576:587], dout1D[672:683], dout2D[0:11], dout2D[96:107], dout2D[192:203], dout2D[288:299], dout2D[384:395], dout2D[480:491], dout2D[576:587], dout2D[672:683], clk, donerow, rst, donecol);
    one_d_idct UC1 (dout1D[12:23], dout1D[108:119], dout1D[204:215], dout1D[300:311], dout1D[396:407], dout1D[492:503], dout1D[588:599], dout1D[684:695], dout2D[12:23], dout2D[108:119], dout2D[204:215], dout2D[300:311], dout2D[396:407], dout2D[492:503], dout2D[588:599], dout2D[684:695], clk, donerow, rst, doneC1);
    one_d_idct UC2 (dout1D[24:35], dout1D[120:131], dout1D[216:227], dout1D[312:323], dout1D[408:419], dout1D[504:515], dout1D[600:611], dout1D[696:707], dout2D[24:35], dout2D[120:131], dout2D[216:227], dout2D[312:323], dout2D[408:419], dout2D[504:515], dout2D[600:611], dout2D[696:707], clk, donerow, rst, doneC2);
    one_d_idct UC3 (dout1D[36:47], dout1D[132:143], dout1D[228:239], dout1D[324:335], dout1D[420:431], dout1D[516:527], dout1D[612:623], dout1D[708:719], dout2D[36:47], dout2D[132:143], dout2D[228:239], dout2D[324:335], dout2D[420:431], dout2D[516:527], dout2D[612:623], dout2D[708:719], clk, donerow, rst, doneC3);
    one_d_idct UC4 (dout1D[48:59], dout1D[144:155], dout1D[240:251], dout1D[336:347], dout1D[432:443], dout1D[528:539], dout1D[624:635], dout1D[720:731], dout2D[48:59], dout2D[144:155], dout2D[240:251], dout2D[336:347], dout2D[432:443], dout2D[528:539], dout2D[624:635], dout2D[720:731], clk, donerow, rst, doneC4);
    one_d_idct UC5 (dout1D[60:71], dout1D[156:167], dout1D[252:263], dout1D[348:359], dout1D[444:455], dout1D[540:551], dout1D[636:647], dout1D[732:743], dout2D[60:71], dout2D[156:167], dout2D[252:263], dout2D[348:359], dout2D[444:455], dout2D[540:551], dout2D[636:647], dout2D[732:743], clk, donerow, rst, doneC5);
    one_d_idct UC6 (dout1D[72:83], dout1D[168:179], dout1D[264:275], dout1D[360:371], dout1D[456:467], dout1D[552:563], dout1D[648:659], dout1D[744:755], dout2D[72:83], dout2D[168:179], dout2D[264:275], dout2D[360:371], dout2D[456:467], dout2D[552:563], dout2D[648:659], dout2D[744:755], clk, donerow, rst, doneC6);
    one_d_idct UC7 (dout1D[84:95], dout1D[180:191], dout1D[276:287], dout1D[372:383], dout1D[468:479], dout1D[564:575], dout1D[660:671], dout1D[756:767], dout2D[84:95], dout2D[180:191], dout2D[276:287], dout2D[372:383], dout2D[468:479], dout2D[564:575], dout2D[660:671], dout2D[756:767], clk, donerow, rst, doneC7);

always @(posedge clk or posedge rst)
begin
   if (rst == 1'b1) 
	begin
	done = 1'b0;
	dout = 768'b0;
	end
   else 
	begin	
	done = donecol;	
	if (donecol == 1'b1) 
		dout = dout2D;
   	end	
end				

endmodule


/******************************************************************************
*
* implements 1-D IDCT 
* This module takes in a row of input frequency, then multilies it by
* the matrix of IDCT coefficients and returns a row of output discrete time
* values.
*         I0, I1, I2 ... I7 are input frequencies.
*         O0, O1, O2 ... O7 are output discrete time values.
* 
******************************************************************************/

module one_d_idct(I0, I1, I2, I3, I4, I5, I6, I7,
                  O0, O1, O2, O3, O4, O5, O6, O7, 
                  clk, go, rst, done);

`define width 14

function [11:0] DropMSBs;
	input [`width+1:0] in;
	if ((in[`width+1:`width-3] == 5'b00000) ||
	    (in[`width+1:`width-3] == 5'b11111))
		DropMSBs = in[`width-3:0];
	else if (in[`width+1] == 1'b0) begin
		DropMSBs = {1'b0,11'b11111111111};
		// $display("Saturating +ve number");
	end
	else begin
		DropMSBs = {1'b1,11'b0};
		// $display("Saturating -ve number");
	end
endfunction

input [11:0] I0;
input [11:0] I1;
input [11:0] I2;
input [11:0] I3;
input [11:0] I4;
input [11:0] I5;
input [11:0] I6;
input [11:0] I7;
input go;
input rst;
output [11:0] O0;
output [11:0] O1;
output [11:0] O2;
output [11:0] O3;
output [11:0] O4;
output [11:0] O5;
output [11:0] O6;
output [11:0] O7;
output done;
wire [11:0] O0;
wire [11:0] O1;
wire [11:0] O2;
wire [11:0] O3;
wire [11:0] O4;
wire [11:0] O5;
wire [11:0] O6;
wire [11:0] O7;
// reg done;
input clk;


reg [`width+1:0] int_O0;
reg [`width+1:0] int_O1;
reg [`width+1:0] int_O2;
reg [`width+1:0] int_O3;
reg [`width+1:0] int_O4;
reg [`width+1:0] int_O5;
reg [`width+1:0] int_O6;
reg [`width+1:0] int_O7;

wire [`width:0] de0, de1, de2, de3;
wire [`width:0] do0, do1, do2, do3;

wire done1, done2;

// Reduce the precision back to 12 bits
assign O0 = DropMSBs(int_O0);
assign O1 = DropMSBs(int_O1);
assign O2 = DropMSBs(int_O2);
assign O3 = DropMSBs(int_O3);
assign O4 = DropMSBs(int_O4);
assign O5 = DropMSBs(int_O5);
assign O6 = DropMSBs(int_O6);
assign O7 = DropMSBs(int_O7);
assign done = done1 & done2;

// O0 is formed by
// C0*I0+C2*I2+C4*I4+C6*I6 + C1*I1+C3*I3+C5*I5+C7*I7
// eqv. to lut0(I0,I2,I4,I6) + lut1(I1,I3,I5,I7)

ByteMultEven multEven (de0, de1, de2, de3, I0, I2, I4, I6, clk, go, rst, done1);
ByteMultOdd multOdd (do0, do1, do2, do3, I1, I3, I5, I7, clk, go, rst, done2);

// Keep 15 bits for the internal computation
always @(de0 or do0 or de1 or do1 or de2 or do2 or de3 or do3)
begin
       int_O0 = {de0[`width],de0[`width:0]} + {do0[`width],do0[`width:0]};
       int_O7 = {de0[`width],de0[`width:0]} - {do0[`width],do0[`width:0]};
       int_O1 = {de1[`width],de1[`width:0]} + {do1[`width],do1[`width:0]};
       int_O6 = {de1[`width],de1[`width:0]} - {do1[`width],do1[`width:0]};
       int_O2 = {de2[`width],de2[`width:0]} + {do2[`width],do2[`width:0]};
       int_O5 = {de2[`width],de2[`width:0]} - {do2[`width],do2[`width:0]};
       int_O3 = {de3[`width],de3[`width:0]} + {do3[`width],do3[`width:0]};
       int_O4 = {de3[`width],de3[`width:0]} - {do3[`width],do3[`width:0]};    
end

endmodule

